@@ -29,9 +29,6 @@ type Client struct {
 
 	// The base URL for requests
 	BaseURL string
-
-	// The internal logger to use
-	Log func(string, ...interface{})
 }
 
 // New creates a new client
@@ -44,12 +41,9 @@ func New(u string) (*Client, error) {
 
 	return &Client{
 		BaseURL: u,
-		Log:     nopLogger,
 	}, nil
 }
 
-var nopLogger = func(_ string, _ ...interface{}) {}
-
 // Validate if the base URL for monocular is valid.
 func validate(u string) error {
 
